import fcntl
import struct
import termios
from contextlib import contextmanager
import tty
import random
import string

max_read_bytes = 1024 * 20


def get_terminal_size(fd):
    data = fcntl.ioctl(fd, termios.TIOCGWINSZ, b"\x00\x00\00\x00")
    rows, cols = struct.unpack("hh", data)
    return rows, cols


def set_terminal_size(fd, rows, cols):
    xpix = ypix = 0
    winsize = struct.pack("HHHH", rows, cols, xpix, ypix)
    fcntl.ioctl(fd, termios.TIOCSWINSZ, winsize)


def copy_terminal_dimensions(src_fd, dest_fd):
    rows, cols = get_terminal_size(src_fd)
    set_terminal_size(dest_fd, rows, cols)


@contextmanager
def make_raw(fd):
    mode = None
    try:
        mode = tty.tcgetattr(fd)
        tty.setraw(fd)
        yield
    except tty.error:
        pass
    except Exception as e:
        print(e)
    if mode:
        tty.tcsetattr(fd, tty.TCSAFLUSH, mode)


def get_random_string(n):
    return "".join(random.choices(string.ascii_letters + string.digits, k=n))



"""
Symmetric encryption with aes gcm
https://cryptography.io/en/latest/hazmat/primitives/aead/#cryptography.hazmat.primitives.ciphers.aead.AESGCM
"""

from cryptography.hazmat.primitives.ciphers.aead import AESGCM  # type: ignore
from cryptography.hazmat.primitives.asymmetric import rsa, padding  # type:ignore
from cryptography.hazmat.primitives import hashes  # type:ignore
from cryptography.hazmat.primitives.serialization import (  # type:ignore
    load_pem_public_key,
)

IV_LENGTH = 12
KEY_LENGTH_BITS = 128


def import_rsa_key(pem_public_key: str):
    pem_bytes = pem_public_key.encode()
    return load_pem_public_key(pem_bytes)


def rsa_encrypt(public_key: rsa.RSAPublicKey, message: bytes) -> bytes:
    return public_key.encrypt(
        message,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )


def aes_generate_secret_key() -> bytes:
    return AESGCM.generate_key(bit_length=KEY_LENGTH_BITS)


def aes_encrypt(message_count: int, key: bytes, data: bytes) -> bytes:
    # the same iv must never be reused with a given key
    iv = message_count.to_bytes(IV_LENGTH, "little")
    # prepend unencrypted iv to the encrypted payload
    return iv + AESGCM(key).encrypt(iv, data, None)


def aes_decrypt(key: bytes, data: bytes) -> str:
    # unencrypted iv must be prepended to the payload
    iv = data[0:IV_LENGTH]
    return AESGCM(key).decrypt(iv, data[IV_LENGTH:], None).decode()



#!/usr/bin/env python3

import argparse
import asyncio
import os
import shlex
from urllib.parse import urlparse
import traceback
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware  # type: ignore
import uvicorn  # type: ignore
from . import share, server

from .constants import TermPairError, TERMPAIR_VERSION

__version__ = TERMPAIR_VERSION


def get_parser():
    p = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="View and control remote terminals from your browser",
    )
    p.add_argument("--version", action="version", version=TERMPAIR_VERSION)
    subparsers = p.add_subparsers(dest="command", required=True)

    share_parser = subparsers.add_parser(
        "share",
        description=(
            "Share your terminal session with one or more browsers. "
            "A termpair server must be running before using this command."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    share_parser.add_argument(
        "--cmd",
        default=os.environ.get("SHELL", "bash"),
        help=(
            "The command to run in this TermPair session. "
            "Defaults to the SHELL environment variable"
        ),
    )
    share_parser.add_argument(
        "--port", "-p", default=8000, help="port server is running on"
    )
    share_parser.add_argument(
        "--host", default="http://localhost", help="host server is running on"
    )
    share_parser.add_argument(
        "--read-only",
        "-r",
        action="store_true",
        help="Do not allow browsers to write to the terminal",
    )
    share_parser.add_argument(
        "--open-browser",
        "-b",
        action="store_true",
        help="Open a browser tab to the terminal after you start sharing",
    )

    server_parser = subparsers.add_parser(
        "serve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Run termpair server to route messages between unix terminals and browsers. "
            "Run this before connecting any clients. "
            "TermPair only works in secure contexts; SSL/TLS is generally required. "
            "To generate an SSL certificate and private key, run "
            "`openssl req -newkey rsa:2048 -nodes -keyout host.key -x509 -days 365 -out host.crt`. "
            "To skip questions and use defaults, add the `-batch` flag. "
            "You can ignore warnings about self-signed certificates since you know you just made it. "
            "Then use them, pass the '--certfile' and '--keyfile' arguments."
        ),
    )
    server_parser.add_argument(
        "--port", "-p", default=8000, help="Port to run the server on"
    )
    server_parser.add_argument(
        "--host",
        default="localhost",
        help="Host to run the server on (0.0.0.0 exposes publicly)",
    )
    server_parser.add_argument(
        "--certfile",
        "-c",
        help="Path to SSL certificate file (commonly .crt extension)",
    )
    server_parser.add_argument(
        "--keyfile",
        "-k",
        help="Path to SSL private key .key file (commonly .key extension)",
    )
    return p


def run_command(args):
    if args.command == "share":
        cmd = shlex.split(args.cmd)

        if not args.host.startswith("http://") and not args.host.startswith("https://"):
            exit("host must start with either http:// or https://")

        parsed = urlparse(args.host)
        if args.port:
            url = f"{parsed.scheme}://{parsed.netloc}:{args.port}{parsed.path}"
        else:
            url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
        url = url if url.endswith("/") else f"{url}/"
        allow_browser_control = not args.read_only
        try:
            asyncio.get_event_loop().run_until_complete(
                share.broadcast_terminal(
                    cmd, url, allow_browser_control, args.open_browser
                )
            )
        except TermPairError as e:
            exit(e)

    elif args.command == "serve":
        if args.certfile or args.keyfile:
            server.app.add_middleware(HTTPSRedirectMiddleware)

        uvicorn.run(
            server.app,
            host=args.host,
            port=int(args.port),
            ssl_certfile=args.certfile,
            ssl_keyfile=args.keyfile,
        )


def main():
    args = get_parser().parse_args()
    try:
        run_command(args)
    except Exception:
        print(
            "TermPair encountered an error. If you think this is a bug, it can be reported at https://github.com/cs01/termpair/issues"
        )
        print("")
        exit(traceback.format_exc())


if __name__ == "__main__":
    main()



from starlette.websockets import WebSocket  # type: ignore

from .Terminal import Terminal


async def handle_ws_message_subprotocol_v3(ws: WebSocket, terminal: Terminal):
    browser_input = await ws.receive_json()
    event = browser_input.get("event")
    if event == "command":
        if terminal.allow_browser_control:
            await terminal.ws.send_json(browser_input)
    else:
        await terminal.ws.send_json(browser_input)



"""
Server to receive all output from user's terminal and forward on to any
browsers that are watching the broadcast
"""


import asyncio
import json
import logging
import os
import time
from hashlib import md5
from typing import Any, Dict, List, Optional

import starlette  # type: ignore
from fastapi import FastAPI  # type: ignore
from starlette.staticfiles import StaticFiles  # type: ignore
from starlette.websockets import WebSocket  # type: ignore
from fastapi.middleware.cors import CORSMiddleware  # type:ignore

from .Terminal import Terminal, TerminalId
from .utils import get_random_string
from .constants import TERMPAIR_VERSION
from .server_websocket_subprotocol_handlers import (
    handle_ws_message_subprotocol_v3,
)
from fastapi.exceptions import HTTPException  # type: ignore


PUBLIC_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "frontend_build")
STATIC_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "frontend_build/static"
)

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=[
        "*",
    ],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

terminals: Dict[TerminalId, Terminal] = {}


@app.get("/ping")
async def ping():
    return "pong"


@app.get("/terminal/{terminal_id}")
async def index(terminal_id: Optional[TerminalId] = None):
    from .main import __version__

    terminal = None
    if terminal_id:
        terminal = terminals.get(terminal_id)

    data: Dict[str, Any]

    if not terminal:
        raise HTTPException(status_code=404, detail="Terminal not found")

    rows = terminal.rows
    cols = terminal.cols
    allow_browser_control = terminal.allow_browser_control
    data = dict(
        terminal_id=terminal_id,
        cols=cols,
        rows=rows,
        allow_browser_control=allow_browser_control,
        command=terminal.command,
        broadcast_start_time_iso=terminal.broadcast_start_time_iso,
        termpair_version=__version__,
    )
    return data


@app.websocket("/connect_browser_to_terminal")
async def connect_browser_to_terminal(ws: WebSocket):
    terminal_id = ws.query_params.get("terminal_id", None)
    terminal = terminals.get(terminal_id)
    if not terminal:
        print(f"terminal id {terminal_id} not found")
        await ws.close()
        return
    await ws.accept()

    terminal.browser_websockets.append(ws)

    # Need to create a task so it can be cancelled by the terminal's
    # task if the terminal session ends. That way, browsers are notified
    # the session ended instead of thinking the connection is still open
    # (an exception raised while awaiting)
    task: asyncio.Task = asyncio.create_task(
        _task_handle_browser_websocket(terminal, ws)
    )

    def remove_task_from_terminal_list(future):
        # task will sit in list as "done"
        # not a big deal if it sits there, but we'll remove it
        # immediately since it's never going to be used again
        terminal.browser_tasks.remove(task)

    task.add_done_callback(remove_task_from_terminal_list)
    terminal.browser_tasks.append(task)
    try:
        # task will be cancelled when terminal session the client started ends
        await task
    except asyncio.exceptions.CancelledError:
        pass


async def _task_handle_browser_websocket(terminal: Terminal, ws: WebSocket):
    try:
        # update connected browser count in each browser
        num_browsers = len(terminal.browser_websockets)
        for browser in terminal.browser_websockets:
            await browser.send_json({"event": "num_clients", "payload": num_browsers})
        while True:
            await handle_ws_message_subprotocol_v3(ws, terminal)

    except starlette.websockets.WebSocketDisconnect:
        # browser closed the connection
        pass
    finally:
        if ws in terminal.browser_websockets:
            terminal.browser_websockets.remove(ws)
        num_browsers = len(terminal.browser_websockets)
        for web_client in terminal.browser_websockets:
            await web_client.send_json(
                {"event": "num_clients", "payload": num_browsers}
            )


async def forward_terminal_data_to_web_clients(terminal: Terminal):
    while True:
        # The task is to endlessly wait for new data from the terminal,
        # read it, and broadcast it to all connected browsers
        ws = terminal.ws
        browser_websockets = terminal.browser_websockets
        try:
            data = await ws.receive_json()
        except starlette.websockets.WebSocketDisconnect:
            # Terminal stopped broadcasting, close
            # all browser websocket tasks so they are notified
            # the connection has actually ended
            for task in terminal.browser_tasks:
                task.cancel()
            return

        terminal_has_closed = False
        event = data.get("event")
        if event == "new_output":
            terminal_data = data.get("payload")
            terminal_has_closed = not terminal_data
        elif event == "resize":
            # namedtuples require you to replace fields
            terminal._replace(rows=data["payload"]["rows"])
            terminal._replace(cols=data["payload"]["cols"])
        elif event in ["aes_keys", "aes_key_rotation"]:
            pass
        else:
            logging.warning(f"Got unknown event {data.get('event', 'none')}")

        if terminal_has_closed:
            # terminal outputs an empty string when it closes, so it just closed
            for browser_ws in browser_websockets:
                # close each browser connection since the terminal's broadcasting
                # process stopped
                await browser_ws.close()
            return

        browsers_to_remove: List[WebSocket] = []
        for browser_ws in browser_websockets:
            try:
                await browser_ws.send_json(data)
            except Exception:
                if browser_ws not in browsers_to_remove:
                    browsers_to_remove.append(browser_ws)

        if browsers_to_remove:
            for browser_ws in browsers_to_remove:
                browser_websockets.remove(browser_ws)

            # let still-connected clients know the new count
            for browser_ws in browser_websockets:
                await browser_ws.send_json(
                    {"event": "num_clients", "payload": len(browser_websockets)}
                )
        # continue running task in while loop


def _gen_terminal_id(ws: WebSocket) -> TerminalId:
    random = str(str(time.time()) + get_random_string(30))
    checksum = md5(random.encode())
    return TerminalId(checksum.hexdigest())


@app.websocket("/connect_to_terminal")
async def connect_to_terminal(ws: WebSocket):
    await ws.accept()
    data = await ws.receive_json()
    subprotocol_version = data.get("subprotocol_version")
    valid_subprotocols = ["3"]
    if subprotocol_version not in valid_subprotocols:
        await ws.send_text(
            json.dumps(
                {
                    "event": "fatal_error",
                    "payload": "Client and server are running incompatible versions. "
                    + f"Server is running v{TERMPAIR_VERSION}. "
                    + "Ensure you are using a version of the TermPair client compatible with the server. ",
                }
            )
        )
        await ws.close()
        return

    terminal_id = _gen_terminal_id(ws)
    terminal = Terminal(
        ws=ws,
        browser_websockets=[],
        browser_tasks=[],
        rows=data["rows"],
        cols=data["cols"],
        allow_browser_control=data["allow_browser_control"],
        command=data["command"],
        broadcast_start_time_iso=data["broadcast_start_time_iso"],
        subprotocol_version=subprotocol_version,
    )
    terminals[terminal_id] = terminal

    # send back to the terminal that the broadcast is starting under
    # this id
    await ws.send_text(json.dumps({"event": "start_broadcast", "payload": terminal_id}))

    # forwards all data from the terminal to browsers for as long as the
    # client is connected
    await asyncio.ensure_future(forward_terminal_data_to_web_clients(terminal))
    terminals.pop(terminal_id, None)


app.mount("/", StaticFiles(directory=PUBLIC_DIR, html=True))
app.mount("/static", StaticFiles(directory=STATIC_DIR, html=True))



import asyncio
from typing import List, NamedTuple, NewType

from starlette.websockets import WebSocket  # type: ignore


class Terminal(NamedTuple):
    ws: WebSocket
    rows: int
    cols: int
    browser_websockets: List[WebSocket]
    browser_tasks: List[asyncio.Task]
    allow_browser_control: bool
    command: str
    broadcast_start_time_iso: str
    subprotocol_version: str


TerminalId = NewType("TerminalId", str)



"""
Establish a websocket connection and replace local terminal with a pty
that sends all output to the server.
"""

import asyncio
import base64
import datetime
import json
import os
import pty
import signal
import ssl
import sys
from urllib.parse import urljoin, urlencode
import webbrowser
from typing import List, Optional, Callable
import websockets  # type: ignore
from math import floor
from .Terminal import TerminalId
from .constants import subprotocol_version, TermPairError
from . import utils
from . import encryption

max_read_bytes = 1024 * 2
ws_queue: asyncio.Queue = asyncio.Queue()
JS_MAX_SAFE_INTEGER = 2 ** 53 - 1


class AesKeys:
    message_count: int

    def __init__(self):
        self.bootstrap_message_count = 0
        self.message_count = 0
        self.message_count_rotation_required = 2 ** 20
        self.browser_rotation_buffer_count = self.message_count_rotation_required * 0.1
        self.secret_bootstrap_key = encryption.aes_generate_secret_key()
        self.secret_unix_key = encryption.aes_generate_secret_key()
        self.secret_browser_key = encryption.aes_generate_secret_key()

    def encrypt_bootstrap(self, plaintext: bytes):
        self.bootstrap_message_count += 1
        return encryption.aes_encrypt(
            self.message_count, self.secret_bootstrap_key, plaintext
        )
        return encryption.aes_encrypt_with_random(self.secret_bootstrap_key, plaintext)

    def encrypt(self, plaintext: bytes):
        self.message_count += 1
        # encrypt with our AES key
        return encryption.aes_encrypt(
            self.message_count, self.secret_unix_key, plaintext
        )

    def decrypt(self, ciphertext: bytes) -> str:
        # decrypt with browser's AES key
        plaintext = encryption.aes_decrypt(self.secret_browser_key, ciphertext)
        return plaintext

    def get_max_iv_for_browser(self, start_iv_count: int) -> int:
        # each browser for this session encrypts using the same AES key.
        # To avoid re-using an IV, we assign each a window to operate within.
        # If the end of the window is hit, a new key is requested.
        max_iv_count = floor(
            start_iv_count
            + self.message_count_rotation_required
            - self.browser_rotation_buffer_count
        )
        if max_iv_count > JS_MAX_SAFE_INTEGER or max_iv_count < start_iv_count:
            raise TermPairError("Cannot create safe AES nonce")
        return max_iv_count

    def get_start_iv_count(self, browser_number: int) -> int:
        start_iv_count = (browser_number - 1) * self.message_count_rotation_required
        if start_iv_count > JS_MAX_SAFE_INTEGER:
            raise TermPairError("Cannot create safe AES nonce")
        return start_iv_count

    @property
    def need_rotation(self) -> bool:
        return self.message_count > self.message_count_rotation_required

    def rotate_keys(self):
        new_unix_key = encryption.aes_generate_secret_key()
        new_browser_key = encryption.aes_generate_secret_key()

        ws_queue.put_nowait(
            json.dumps(
                {
                    "event": "aes_key_rotation",
                    "payload": {
                        "b64_aes_secret_unix_key": base64.b64encode(
                            self.encrypt(new_unix_key)
                        ).decode(),
                        "b64_aes_secret_browser_key": base64.b64encode(
                            self.encrypt(new_browser_key)
                        ).decode(),
                    },
                }
            )
        )
        self.secret_unix_key = new_unix_key
        self.secret_browser_key = new_browser_key
        self.message_count = 0


class SharingSession:
    stdout_fd: int
    num_browsers: int

    def __init__(
        self,
        url: str,
        cmd: List[str],
        pty_fd: int,
        stdin_fd: int,
        stdout_fd: int,
        ws,
        open_browser: bool,
        allow_browser_control: bool,
    ):
        self.url = url
        self.cmd = cmd
        self.pty_fd = pty_fd
        self.stdin_fd = stdin_fd
        self.stdout_fd = stdout_fd
        self.ws = ws
        self.open_browser = open_browser
        self.allow_browser_control = allow_browser_control
        self.aes_keys = AesKeys()
        self.terminal_id = None
        self.num_browsers = 0

    async def register_broadcast_with_server(self) -> TerminalId:
        """Prepare server to store i/o about this terminal"""
        # copy our terminal dimensions to the pty so its row/col count
        # matches and we don't get unexpected line breaks/misalignments
        utils.copy_terminal_dimensions(self.stdin_fd, self.pty_fd)

        rows, cols = utils.get_terminal_size(self.stdin_fd)

        cmd_str = " ".join(self.cmd)
        broadcast_start_time_iso = datetime.datetime.now(
            datetime.timezone.utc
        ).isoformat()
        await self.ws.send(
            json.dumps(
                {
                    "rows": rows,
                    "cols": cols,
                    "allow_browser_control": self.allow_browser_control,
                    "command": cmd_str,
                    "broadcast_start_time_iso": broadcast_start_time_iso,
                    "subprotocol_version": subprotocol_version,
                }
            )
        )
        event, payload = await self.receive_data_from_websocket()
        if event == "start_broadcast":
            return payload
        elif event == "fatal_error":
            raise TermPairError(fatal_server_error_msg(payload))
        else:
            raise TermPairError(
                "Unexpected event type received when starting broadcast. "
                + "Ensure you are using a compatible version with the server.",
                event,
            )

    async def run(self):
        self.terminal_id = await self.register_broadcast_with_server()

        self.print_broadcast_init_message()

        with utils.make_raw(self.stdin_fd):
            await self.do_broadcast()

    async def do_broadcast(self):
        signal.signal(
            signal.SIGWINCH,
            lambda signum, frame: self.emit_terminal_dimensions(),
        )
        if self.open_browser:
            webbrowser.open(self.share_url)

        tasks = [
            asyncio.ensure_future(self.task_send_ws_queue_to_server()),
            asyncio.ensure_future(self.task_receive_websocket_messages()),
        ]

        loop = asyncio.get_event_loop()

        def cleanup():
            for t in tasks:
                t.cancel()
                loop.remove_reader(self.stdin_fd)
                loop.remove_reader(self.pty_fd)

        # add event-based reading of input to stdin, and forward to the pty
        # process
        loop.add_reader(self.stdin_fd, self.handle_new_stdin)

        # add event based reading of output from the pty and write to
        # stdout and to the server
        loop.add_reader(
            self.pty_fd,
            self.handle_new_pty_output,
            cleanup,
        )

        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()

    def handle_new_stdin(self):
        """forwards from terminal's stdin to the pty's stdin"""
        user_input = os.read(self.stdin_fd, max_read_bytes)
        os.write(self.pty_fd, user_input)

    def emit_terminal_dimensions(self):
        utils.copy_terminal_dimensions(self.stdin_fd, self.pty_fd)
        rows, cols = utils.get_terminal_size(self.stdin_fd)
        ws_queue.put_nowait(
            json.dumps({"event": "resize", "payload": {"rows": rows, "cols": cols}})
        )

    async def task_receive_websocket_messages(self):
        """receives events+payloads from browser websocket connection"""
        try:
            while True:
                event, payload = await self.receive_data_from_websocket()
                if event == "command":
                    if self.allow_browser_control:
                        try:
                            encrypted_payload = base64.b64decode(payload)
                            json_data = self.aes_keys.decrypt(encrypted_payload)
                            data = json.loads(json_data)["data"]
                            os.write(self.pty_fd, data.encode())
                        except Exception:
                            pass
                elif event == "request_terminal_dimensions":
                    self.emit_terminal_dimensions()
                elif event == "request_key_rotation":
                    self.aes_keys.rotate_keys()
                elif event == "new_browser_connected":
                    # no payload required, just the event.
                    # the browser must already have
                    # the bootstrap AES key to decrypt the
                    # message we're about to
                    self.num_browsers += 1
                    try:
                        # use the bootstrap AES key to encrypt the two AES keys
                        # used for terminal input/output
                        b64_bootstrap_unix_aes_key = base64.b64encode(
                            self.aes_keys.encrypt_bootstrap(
                                self.aes_keys.secret_unix_key
                            )
                        ).decode()
                        b64_bootstrap_browser_aes_key = base64.b64encode(
                            self.aes_keys.encrypt_bootstrap(
                                self.aes_keys.secret_browser_key
                            )
                        ).decode()

                        iv_count = self.aes_keys.get_start_iv_count(self.num_browsers)
                        ws_queue.put_nowait(
                            json.dumps(
                                {
                                    "event": "aes_keys",
                                    "payload": {
                                        "b64_bootstrap_unix_aes_key": b64_bootstrap_unix_aes_key,
                                        "b64_bootstrap_browser_aes_key": b64_bootstrap_browser_aes_key,
                                        "iv_count": iv_count,
                                        "max_iv_count": self.aes_keys.get_max_iv_for_browser(
                                            iv_count
                                        ),
                                        "salt": base64.b64encode(
                                            os.urandom(12)
                                        ).decode(),
                                    },
                                }
                            )
                        )
                    except Exception as e:
                        print(e)
                elif event == "fatal_error":
                    raise fatal_server_error_msg(payload)
                else:
                    # TODO log to a file
                    pass
        except websockets.exceptions.ConnectionClosed:
            return

    async def task_send_ws_queue_to_server(self):
        """Waits for new pty output (nonblocking), then immediately sends to server"""
        while True:
            data = await ws_queue.get()
            await self.ws.send(data)

    async def receive_data_from_websocket(self):
        data = await self.ws.recv()
        parsed = json.loads(data)
        return parsed["event"], parsed.get("payload")

    def print_broadcast_init_message(self):
        _, cols = utils.get_terminal_size(sys.stdin)
        secret_key_b64 = base64.b64encode(self.aes_keys.secret_bootstrap_key).decode()
        msg = [
            "\033[1m\033[0;32mConnection established with end-to-end encryption\033[0m 🔒",
            "",
            "Shareable link: "
            + self.get_share_url(self.url, self.terminal_id, secret_key_b64),
            "",
            f"Terminal ID: {self.terminal_id}",
            f"Secret encryption key: {secret_key_b64}",
            f"TermPair Server URL: {self.url}",
            "",
            "Type 'exit' or close terminal to stop sharing.",
        ]

        dashes = "-" * cols
        print(dashes)
        for m in msg:
            print(m)
        print(dashes)

    def get_share_url(
        self,
        url: str,
        terminal_id: str,
        secret_key_b64: str,
    ):
        # if static_url:
        #     qp = {
        #         "terminal_id": terminal_id,
        #         "termpair_server_url": url,
        #     }
        #     return urljoin(static_url, f"?{urlencode(qp)}")
        # else:

        qp = {
            "terminal_id": terminal_id,
        }
        return urljoin(url, f"?{urlencode(qp)}#{secret_key_b64}")

    def handle_new_pty_output(self, cleanup: Callable):
        """forwards pty's output to local stdout AND to websocket"""
        try:
            pty_output = os.read(self.pty_fd, max_read_bytes)
        except OSError:
            cleanup()
            return

        if pty_output:
            # forward output to user's terminal
            os.write(self.stdout_fd, pty_output)

            # also forward output to the server so it can forward to connected browsers
            plaintext_payload = json.dumps(
                {
                    "pty_output": base64.b64encode(pty_output).decode(),
                    "salt": base64.b64encode(os.urandom(12)).decode("utf8"),
                }
            )
            encrypted_payload = self.aes_keys.encrypt(plaintext_payload.encode())
            ws_queue.put_nowait(
                json.dumps(
                    {
                        "event": "new_output",
                        "payload": base64.b64encode(encrypted_payload).decode(),
                    }
                )
            )
            if self.aes_keys.need_rotation:
                self.aes_keys.rotate_keys()
        else:
            cleanup()
            return


def fatal_server_error_msg(error_msg: str):
    raise TermPairError("Connection was terminated with a fatal error: " + error_msg)


async def broadcast_terminal(
    cmd: List[str], url: str, allow_browser_control: bool, open_browser: bool
):
    """Fork this process and connect it to websocket to broadcast it"""
    # create child process attached to a pty we can read from and write to

    (child_pid, pty_fd) = pty.fork()
    if child_pid == 0:
        # This is the forked process. Replace it with the shell command
        # the user wants to run.
        env = os.environ.copy()
        env["TERMPAIR_BROADCASTING"] = "1"
        env["TERMPAIR_BROWSERS_CAN_CONTROL"] = "1" if allow_browser_control else "0"
        os.execvpe(cmd[0], cmd, env)
        return

    stdin_fd = sys.stdin.fileno()
    stdout_fd = sys.stdout.fileno()

    ssl_context: Optional[ssl.SSLContext] = (
        ssl.SSLContext(ssl.PROTOCOL_TLS) if url.startswith("https") else None
    )

    ws_url = url.replace("http", "ws")

    ws_endpoint = urljoin(ws_url, "connect_to_terminal")
    try:
        async with websockets.connect(ws_endpoint, ssl=ssl_context) as ws:
            sharing_session = SharingSession(
                url,
                cmd,
                pty_fd,
                stdin_fd,
                stdout_fd,
                ws,
                open_browser,
                allow_browser_control,
            )
            await sharing_session.run()
            print(
                f"You are no longer broadcasting terminal id {sharing_session.terminal_id}"
            )
    except ConnectionRefusedError as e:
        raise TermPairError(
            "Connection was refused. Is the TermPair server running on the host and port specified? "
            + str(e),
        )



#!/usr/bin/env python

import ast
import io
import re
import os
from setuptools import find_packages, setup  # type: ignore
import distutils.text_file

EXCLUDE_FROM_PACKAGES = ["contrib", "docs", "tests*"]
CURDIR = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(CURDIR, "README.md"), "r", encoding="utf-8") as f:
    README = f.read()


def get_version():
    main_file = os.path.join(CURDIR, "termpair", "constants.py")
    _version_re = re.compile(r"TERMPAIR_VERSION\s+=\s+(?P<version>.*)")
    with open(main_file, "r", encoding="utf8") as f:
        match = _version_re.search(f.read())
        version = match.group("version") if match is not None else '"unknown"'
    return str(ast.literal_eval(version))


setup(
    name="termpair",
    version=get_version(),
    author="Chad Smith",
    author_email="chadsmith.software@gmail.com",
    description="View and control remote terminals from your browser with end-to-end encryption",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/cs01/termpair",
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    include_package_data=True,
    keywords=["e2ee", "secure", "terminal", "share", "broadcast", "pty", "websockets"],
    scripts=[],
    entry_points={"console_scripts": ["termpair=termpair.main:main"]},
    zip_safe=False,
    install_requires=distutils.text_file.TextFile(
        filename="./requirements.txt"
    ).readlines(),
    python_requires=">=3.6",
    # license and classifier list:
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    license="License :: OSI Approved :: MIT License",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 4 - Beta",
    ],
)
